"""
This file is part of famafrench.
Copyright (c) 2020, Christian Jauregui <chris.jauregui@berkeley.edu>
See file LICENSE.txt for license information.

Filename
_________
`famafrench/utils.py`

Descriptions
____________
lru_cached_method
    Wrapper for methodtools.lru_cache(maxsize) that allows for sphinx documentation
    recognition of decorated instance methods.
get_kfpriorfactors_directly
    Directly download zipped datafiles from Ken French's online data library without the
    use of the 'pandas-datareader' package. This method is used for a select few
    number of datafiles for which 'pandas-datareader' returns an error.
timing
    Wrapper for class methods that are to be timed for speed and performance measurement.
any_in
    Provide a boolean variable that is =True if elements in a given set intersect with
    elements in another set, =False, otherwise.
priormonthToDay
    Using Fama and French's methodology, map the prior (j-k) monthly return strategy into
    a daily strategy (see online documentation provided on Ken French's website).
grouped_vwAvg
     Calculate weighted (net) portfolio return for a given portfolio with weights within a
     group or set of groups. This function is FASTER THAN groupby(...).apply(...) because it
     avoids non-optimized aggregation.
portRetAvg
    Compute a simple average across different columns.
get_statsTable
    Construct tables with formatted summary statistics.
"""

__author__ = 'Christian Jauregui <chris.jauregui@berkeley.edu'
__all__ = [
    "lru_cached_method",
    "get_kfpriorfactors_directly",
    "timing",
    "any_in",
    "priormonthToDay",
    "grouped_vwAvg",
    "portRetAvg",
    "get_statsTable",
]

# Standard Imports
import weakref
import pandas as pd
import numpy as np
from functools import wraps
from methodtools import lru_cache  # see documentation: https://pypi.org/project/methodtools/
from time import time

# Function: lru_cached_method(.,.):
def lru_cached_method(*lru_args, **lru_kwargs):
    """
    Wrapper for :func:`methodtools.lru_cache` enabling recognition of `decorated`
    class instance methods by `Sphinx <https://www.sphinx-doc.org/en/master/>`_.

    Parameters
    __________
    *lru_args : arbitrary
        Variable number of arguments to :func:`methodtools.lru_cache`

    **lru_kwargs : arbitrary
        Keyworded, variable-length argument list for :func:`methodtools.lru_cache`

    Returns
    _______
    decorator : arbitrary `wrapped` object
        Wrapped function.
    """
    def decorator(wrapped_fn):
        @wraps(wrapped_fn)
        def wrapped(self, *args, **kwargs):
            # Use a weak reference to self. This prevents a self-reference cycle that fools the garbage collector
            # into thinking the instance shouldn't be dropped when all external references are dropped.
            weak_ref_to_self = weakref.ref(self)
            @wraps(wrapped_fn)
            @lru_cache(*lru_args, **lru_kwargs)
            def cached(*args, **kwargs):
                return wrapped_fn(weak_ref_to_self(), *args, **kwargs)
            setattr(self, wrapped_fn.__name__, cached)
            return cached(*args, **kwargs)
        return wrapped
    return decorator


# Function: get_kfpriorfactors_directly(.,.):
def get_kfpriorfactors_directly(kflib_name, kflib_freq, kf_factor):
    """
    Directly download (from Ken French's online library) zipped monthly or annual datafiles
    for the `Short-Term Reversal` or `Long-Term Reversal` Fama-French-style factors.
    This is required since the :meth:`pandas_datareader.web` method is broken for such datafiles.

    Parameters
    ___________
    kflib_name : str
        Name of zipped datafile.
    kflib_freq : str
        Observation frequency of factor portfolios. Possible choices are:

            * ``M``: monthly
            * ``A``: annual
    kf_factor : str
        The name or "label" of the Fama-French-style factor. Possible choices are:

            * ``ST_Rev`` : Short-Term Reversal - based on Prior (1-1) returns
            * ``LT_Rev`` : Long-Term Reversal - based on Prior (13-60) returns

    Returns
    ________
    kflib_data : pandas.DataFrame
        Dataset with time-series of the Fama-French-style factor.
    """
    from io import BytesIO
    from zipfile import ZipFile
    from urllib.request import urlopen

    urllink = "http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/"+kflib_name+"_CSV.zip"
    url = urlopen(urllink)

    # Download Zipfile and create pandas.DataFrame
    zipfile = ZipFile(BytesIO(url.read()))
    kflib_data = pd.read_csv(zipfile.open(kflib_name+".CSV"), header=0, skiprows=13)
    kflib_data.rename(columns={'Unnamed: 0': 'Date'}, inplace=True)
    kflib_data.columns = list(map(str.rstrip, kflib_data.columns))

    if kf_factor not in ['ST_Rev', 'LT_Rev']:
        kf_factor = kf_factor.title()

    # Get first row number that has all values set to "nan" coinciding with the start of annual data
    rows_with_all_nan = kflib_data.index[kflib_data[kf_factor].isna()]
    diffrows_with_all_nan = [j-i for i, j in zip(rows_with_all_nan[:-1], rows_with_all_nan[1:])]
    first_nan = rows_with_all_nan[0]
    second_nan = rows_with_all_nan[diffrows_with_all_nan.index(max(diffrows_with_all_nan))+1]

    if kflib_freq == 'M':
        kflib_data = kflib_data[:first_nan]
        kflib_data.loc[:, 'Date'] = pd.to_datetime(kflib_data['Date'], format='%Y%m').dt.to_period('M')
    elif kflib_freq == 'A':
        kflib_data = kflib_data[first_nan+4:second_nan]
        kflib_data.loc[:, 'Date'] = pd.to_datetime(kflib_data['Date'], format='%Y').dt.to_period('Y')
    kflib_data = kflib_data[['Date', kf_factor]]
    kflib_data[kf_factor] = kflib_data[kf_factor].replace(-99.99, np.nan).astype(float)
    if kf_factor not in ['ST_Rev', 'LT_Rev']:
        kflib_data = kflib_data.rename(columns={kf_factor: kf_factor.title().upper()})
    kflib_data.set_index('Date', inplace=True)
    return kflib_data


# Function: timing(.)
def timing(func):
    """
    Wrapper for class instance methods enabling the timing of execution.
    Important for measuring speed and performance measurement.

    Parameters
    ___________
    func : func
        Function to be wrapped and timed following execution.

    Returns
    ________
    wrapper : arbitrary `wrapped` object
        The wrapped result(s) for the function `func`.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        t0 = time()
        result = func(*args, **kwargs)
        print("Elapsed time: ", round(time() - t0, 3), " seconds.\n")
        return result
    return wrapper


# Function: any_in(.,.)
def any_in(a_set, b_set):
    """
    Boolean variable that is ``True`` if elements in a given set `a_set` intersect
    with elements in another set `b_set`. Otherwise, the boolean is ``False``.

    Parameters
    ___________
    a_set : list
        First set of elements.
    b_set : list
        Second set of elements.

    Returns
    ________
    not set(a_set).isdisjoint(b_set) : bool
        Boolean that is ``True`` if there is a non-empty intersection between both sets.
        Otherwise, the boolean is ``False``.
    """
    return not set(a_set).isdisjoint(b_set)


# Function: priormonthToDay(.,.,.)
def priormonthToDay(freq, j_mth, k_mth):
    """
    Consistent w/ Fama and French (2008, 2016), map the prior `(j-k)` monthly return strategy into a daily strategy
    (see `Ken French's online documentation  <https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html>`_).

    Parameters
    ___________
    freq : str
        Frequency used to calculate prior `(j-k)` return strategy. Possible choices are:

            * ``D`` : daily
            * ``'M`` : monthly
    j_mth : str
        Lagged month (or day) we start measuring stock performance.
    k_mth : str
        How many months (or days) are used in measuring stock performance.

    Returns
    ________
    j_per, k_per : tuple, str
        ``j_per = j_mth`` and ``k_per = k_mth`` if ``freq = M``.
        Otherwise, monthly figures mapped to daily periods using the description found on Ken French's online documentation:

            * `Daily Momentum <https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_mom_factor_daily.html>`_.
            * `Daily Short-Term Reversal <https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_st_rev_factor_daily.html>`_.
            * `Daily Long-Term Reversal <https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_lt_rev_factor_daily.html>`_.

    Note
    ____
    Monthly ``M`` (daily ``D``) strategies involve portfolios formed every month `t-1` (or day `t-1`)
    for month `t` (or day `t`).

    Note
    ____
    The Fama and French (2008, 2016) momentum strategy definition differs from that of Jegadeesh and Titman (1993).
    Jegadeesh and Titman (1993) consider **J/K** strategies, which include portfolios formed on stock performance over the previous
    **J** months (excluding the last week or month prior to portfolio formation, to remove the
    large short-horizon reversals associated with bid-ask bounce) and hold portfolios for **K** months, where **J**, **K** :math:`\in` {3,6,9,12}.
    Future updates to this module will extend this package to include these additional momentum strategies.


    References
    ___________
    *   Fama, Eugene F., and Kenneth R. French. (2008). `Dissecting Anomalies`,
        Journal of Finance, 48(4), pp.1653-1678

    *   Fama, Eugene F., and Kenneth R. French. (2016). `Dissecting Anomalies with a Five-Factor Model`,
        Journal of Finance, 48(4), pp.1653-1678
    """
    # Per Fama and French, map prior (2-12), (1-1), and (13-60) returns
    # at the monthly frequency to the daily frequency
    if freq in ['D', 'W']:
        if (j_mth == '2') and (k_mth == '12'):
            j_per, k_per = '21', '250'
            return j_per, k_per
        elif (j_mth == '1') and (k_mth == '1'):
            j_per, k_per = '1', '20'
            return j_per, k_per
        elif (j_mth == '13') and (k_mth == '60'):
            j_per, k_per = '251', '1250'
            return j_per, k_per
        else:
            raise ValueError('\'prior (j-k)\' return strategy not of the standard Fama and French type.')
    elif freq in ['M', 'Q', 'A']:
        j_per, k_per = j_mth, k_mth
        return j_per, k_per
    else:
        raise ValueError('Please specify one of the following frequencies: \'D\' or \'M\'')


# Function: grouped_vwAvg(.,.,.,.,.)
def grouped_vwAvg(df0, col_values, col_weights, *groupby_args, **groupby_kwargs):
    """
    Calculate (net) weighted portfolio return for portfolio with weights ``col_weights`` within a group or groups.

    Parameters
    ___________
    df0 : pandas.DataFrame
        Dataset containing firm-level stock returns and portfolio weights, both indexed by dates and firm identifiers when required.
    col_values : list, str
        Column(s) to average over.
    col_weights : str
        Column containing the portfolio weights.
    group_args : list, str, [optional]
        args to pass into `groupby` (ie the level to group on).
    group_kwargs : list, str, [optional]
        kwargs to pass into `groupby`.

    Returns
    ________
    df1 : pandas.Series, or pandas.DataFrame
        Original dataset augmented w/ value-weighted returns.

    Note
    ____
    Function is **FASTER** than :func:`pandas.core.groupby.GroupBy.apply` since it avoids non-optimized aggregation.
    """
    if isinstance(col_values, str):
        col_values = [col_values]

    ss = []
    for value in col_values:
        df1 = df0.copy()
        value_x_weight = 'product_{v}_{w}'.format(v=value, w=col_weights)
        weights = 'weights_{w}'.format(w=col_weights)

        df1[value_x_weight] = df1[value] * df1[col_weights]
        df1[weights] = df1[col_weights].where(~df1[value_x_weight].isnull())
        df1 = df1.groupby(*groupby_args, **groupby_kwargs).sum()
        s = df1[value_x_weight] / df1[weights]

        s.name = value
        ss.append(s)
    df1 = pd.concat(ss, axis=1) if len(ss) > 1 else ss[0]
    return df1


# Function: portRetAvg(.)
def portRetAvg(df):
    """
    Compute a simple average across different columns.

    Parameters
    ___________
    df : pandas.DataFrame
        Dataset with columns to average over.

    Returns
    ________
    dfavg : pandas.DataFrame
        Dataset with the averaged columns.
    """
    dfavg = df.sum(axis=1, min_count=len(df.columns)) / len(df.columns)
    return dfavg


# Function: get_statsTable(.,.,.,.)
def get_statsTable(dType, dFreq, df, dates_as_index=True, ptiles=None):
    """
    Construct detailed tables with summary statistics.

    Parameters
    ___________
    dType : str
        Dataset type of the portfolios. Possible choices are:

            * ``Returns``
            * ``Factors``
            * ``NumFirms``
            * ``Characs``
    dFreq : str
        Observation frequency of the portfolios. Possible choices are:

            * ``D`` : daily
            * ``W`` : weekly
            * ``M`` : monthly
            * ``Q`` : quarterly (3-months)
            * ``A`` : annual
    df : pandas.DataFrame
        Dataset w/ portfolio returns (which may include factor returns), number of firms in each portfolio,
        or `average` anomaly portfolio characteristics for a given portfolio sorting strategy.
    dates_as_index : bool
        Flag determining whether ``df`` has a :class:`pandas.DatetimeIndex` index (``dates_as_index = True``).
        Otherwise, ``dates_as_index = False``.
    ptiles : list, float, default None
        List of percentiles (in decimal format) included as part of output results.
        If ``None``, then ``ptiles = [0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99]``.


    Returns
    ________
    statsTable : pandas.DataFrame
        Summary statistics of the dataset including the following:

            * number of observations :meth:`pandas.DataFrame.count`
            * sample mean :meth:`pandas.DataFrame.mean`
            * sample standard deviation :meth:`pandas.DataFrame.std`
            * sample min :meth:`pandas.DataFrame.min`
            * sample max :meth:`pandas.DataFrame.max`
            * sample skewness :meth:`pandas.DataFrame.skew`
            * sample kurtosis :meth:`pandas.DataFrame.kurtosis`
            * sample mean absolute deviation :meth:`pandas.DataFrame.mad`
            * sample percentiles :meth:`numpy.percentile`

        If ``dates_as_index = True``, then the table also includes the starting and ending date for each observation type.
    """
    if ptiles is None:
        ptiles = [0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99]
    statsTable = df.describe(percentiles=ptiles).round(2)
    statsTable = statsTable.append(df.reindex(statsTable.columns, axis=1).agg(['skew', 'kurt', 'mad'])).round(2)

    # Find starting and ending date of each observation type
    if dates_as_index:
        statsTable0 = pd.DataFrame(index=['startdate', 'enddate'], columns=df.columns)
        for col in df.columns:
            sdate = df[col].first_valid_index()
            edate = df[col].index.max()
            statsTable0.loc['startdate', col] = sdate
            statsTable0.loc['enddate', col] = edate
        statsTable = statsTable0.append(statsTable)

    if dType in ['Factors', 'Returns']:
        if dates_as_index:
            statsTable.iloc[3:, :] = statsTable.iloc[3:, :].astype(str) + '%'
        else:
            statsTable.iloc[1:, :] = statsTable.iloc[1:, :].astype(str) + '%'
        statsTable.loc['count', :] = statsTable.loc['count', :].apply('{:.0f}'.format)
    elif dType == 'NumFirms':
        for c in statsTable.columns:
            if dates_as_index:
                statsTable.loc[2:, c] = statsTable.loc[:, c].apply('{:.0f}'.format)
            else:
                statsTable.loc[:, c] = statsTable.loc[:, c].apply('{:.0f}'.format)
    else:
        statsTable.loc['count', :] = statsTable.loc['count', :].apply('{:.0f}'.format)
    print('    *********************** Observation frequency: '+dFreq+' ************************')
    print(statsTable, '\n')
    return statsTable


"""
This file is part of famafrench.
Copyright (c) 2020, Christian Jauregui <chris.jauregui@berkeley.edu>
See file LICENSE.txt for license information.

Filename
________
`setup.py`

Note
_____
setuptools-based setup module:
* see https://packaging.python.org/guides/distributing-packages-using-setuptools/
* see https://setuptools.readthedocs.io/en/latest/setuptools.html

Note
_____
For more information on licenses, go to https://choosealicense.com/licenses/
"""
from os import path
from setuptools import setup, find_packages

# Define __version__
exec(open("famafrench/version.py").read())

if '__version__' in locals():
    VERSION = __version__
else:
    VERSION = '0.1.4'

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.md'), "r") as readme:
    long_description = readme.read()

NAME = 'famafrench'

packages = ['famafrench']

REQUIRES = [
    'IPython >= 7.12.0',
    'numpy >= 1.16.1',
    'numpydoc >=0.9.2',
    'numba >= 0.48.0',
    'methodtools >= 0.1.0',
    'pandas >= 0.24.2',
    'pandas_datareader >= 0.7.0',
    'pandas_market_calendars >= 1.1',
    'python_dateutil >= 2.8.1',
    'python_dotenv >= 0.13.0',
    'sqlalchemy >= 1.3.13',
    'sphinx >= 2.0',
    'termcolor >= 1.1.0',
    'tqdm >= 4.41.1',
    'wrds >= 3.0.8'
]

setup(
    # Package meta-data
    name=NAME,
    version=VERSION,
    author='Christian Jauregui',
    author_email='chris.jauregui@berkeley.edu',
    description='Python package designed to construct and replicate datasets from Ken French\'s '
                'online library (https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html) '
                'via remote access to wrds-cloud. ',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/christianjauregui/famafrench',
    license='Apache License 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    # See https://packaging.python.org/tutorials/packaging-projects/#classifiers
    # https://www.python.org/dev/peps/pep-0440/
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Office/Business :: Financial',
        'Topic :: Office/Business :: Financial :: Investment',
        "Topic :: Scientific/Engineering",
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],

    # What does your project relate to?
    keywords=[
        'alpha',
        'alpha-factors',
        'beta',
        'capm',
        'econometrics',
        'equities',
        'equity-returns',
        'expected-returns'
        'factor-investing',
        'factor-model',
        'factor-returns',
        'fama-french',
        'finance',
        'growth',
        'investment',
        'momentum',
        'portfolio',
        'risk-model',
        'securities',
        'stock-returns',
        'value',
    ],

    project_urls={
        'Documentation': 'https://christianjauregui.github.io/famafrench/',
        'Source': 'https://github.com/christianjauregui/famafrench',
        'Tracker': 'https://github.com/christianjauregui/famafrench/issues',
    },

    # Requirements
    packages=find_packages(exclude=['contrib',
                                    'docs',
                                    'tests*',
                                    '*txt',
                                    '*md',
                                    '*rst']),
    install_requires=REQUIRES,
    include_package_data=True,
    python_requires='>=3.4.0',
    tests_require=['pytest'],
    test_suite="tests"
    )


"""
This file is part of famafrench.
Copyright (c) 2020, Christian Jauregui <chris.jauregui@berkeley.edu>
See file LICENSE.txt for license information.

This file must be kept very simple, because it is consumed from several places.
It is imported by famafrench/__init__.py, execfile'd by setup.py, etc.

I use a simple scheme: 1.0.0 -> 1.0.0+dev -> 1.1.0 -> 1.1.0+dev
where the +dev versions are never publicly released - they are just what I
stick into the VCS in between releases.

This is compatible with PEP 440:
   http://legacy.python.org/dev/peps/pep-0440/
via the use of the "local suffix" "+dev", which is disallowed on index
servers and causes 1.0.0+dev to sort after plain 1.0.0, which is what I
want. (Contrast with the special suffix 1.0.0.dev, which sorts *before* 1.0.0.)

"""

__version__ = "0.1.4"



"""
This file is part of famafrench.
Copyright (c) 2020, Christian Jauregui <chris.jauregui@berkeley.edu>
See file LICENSE.txt for license information.

Filename
________
`examples/famafrench_priorreturnsfactors.py`

Construct Fama/French Factors based on Prior Returns:
* Momentum: prior (2-12) returns (MOM)
* Short-Term Reversal: prior (1-1) returns (ST_Rev)
* Long-Term Reversal: prior (13-60) returns (LT_Rev)

Ken French online library:
Momentum
________
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_mom_factor.html
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_mom_factor_daily.html

Short-Term Reversal
___________________
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_st_rev_factor.html
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_st_rev_factor_daily.html

Long-Term Reversal
__________________
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_lt_rev_factor.html
https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/Data_Library/det_lt_rev_factor_daily.html

"""
import os
import datetime as dt
import famafrench.famafrench as ff

startDate = dt.date(1960, 1, 1)  # "default" startDate
endDate = dt.date.today()  # "default" endDate
#startDate = dt.date(1970, 1, 1)
#endDate = dt.date(2019, 12, 31)

ffsortCharac = ['ME', 'PRIOR_2_12']
ffFactors = ['MKT-RF', 'MOM', 'ST_Rev', 'LT_Rev']
ffportCharac = ['ME', 'PRIOR_2_12', 'PRIOR_1_1', 'PRIOR_13_60']

# pickled_dir
pickled_dir = os.getcwd() + '/famafrench/pickled_db/'

#%%
#********************************************************************************************#
# Example 1: daily sample
#********************************************************************************************#
runQuery = True
ffFreq = 'D'
ff_D = ff.FamaFrench(pickled_dir, runQuery, ffFreq, ffsortCharac, ffFactors, ffportCharac)

# Summary statistics
ff_D.getFamaFrenchStats('Factors', ffFreq, startDate, endDate)

# Compare daily Fama-French factors based on prior returns constructed here to those provided in Ken French's online library
portTableD = ff_D.getFFfactors(startDate, endDate)
kfportTableD = ff_D.getkfFFfactors(ffFreq, startDate, endDate)
_, _, _, = ff_D.comparePortfolios('Factors', ffFreq, startDate, endDate)


#%%
#********************************************************************************************#
# Example 2: monthly sample
#********************************************************************************************#
runQuery = True
ffFreq = 'M'
ff_M = ff.FamaFrench(pickled_dir, runQuery, ffFreq, ffsortCharac, ffFactors, ffportCharac)

# Summary statistics
ff_M.getFamaFrenchStats('Factors', ffFreq, startDate, endDate)

# Compare monthly Fama-French factors based on prior returns constructed here to those provided in Ken French's online library
portTableM = ff_M.getFFfactors(startDate, endDate)
kfportTableM = ff_M.getkfFFfactors(ffFreq, startDate, endDate)
_, _, _, = ff_M.comparePortfolios('Factors', ffFreq, startDate, endDate)


#%%
#********************************************************************************************#
# Example 3: annual sample
#********************************************************************************************#
runQuery = False
ffFreq = 'A'
ff_A = ff.FamaFrench(pickled_dir, runQuery, ffFreq, ffsortCharac, ffFactors, ffportCharac)

# Summary statistics
ff_A.getFamaFrenchStats('Factors', ffFreq, startDate, endDate)

# Compare annual Fama-French factors based on prior returns constructed here to those provided in Ken French's online library
_, _, _, = ff_A.comparePortfolios('Factors', ffFreq, startDate, endDate)




